/* -*-Asm-*- */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2000 Machine Vision Holdings, Inc.
 *
 *  Author: David Woodhouse <dwmw2@infradead.org>
 *
 *  $Id: doc_stage1.S,v 1.6 2000/01/31 09:49:31 dwmw2 Exp $
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "doc_stage1.h"
#include "stage2_size.h"

/*
 *  defines for the code go here
 */

#ifdef OLDGRUB /* Loading Erich's old grub */

#define LOADLEN (((STAGE2_SIZE + 0xff) & ~0xff) >> 8)
#define GRUBSTART 0x8000
#define GRUBLOADSEG	0x7e0
#define GRUBLOADOFS	0x200

#else /* Loading FSF grub */

#define LOADLEN (((((STAGE2_SIZE + 0xff) & ~0xff) + 0x200)) >> 8)
#define GRUBSTART 0x8200
#define GRUBLOADSEG	0x800
#define GRUBLOADOFS	0x000

#endif


	
			/* Some macros to make it obvious what we're accessing */
#define BXREG 0x1004

#define BX_ChipID		-4(%bx)
#define BX_DOCControl		-2(%bx)
#define BX_CSDNControl		(%bx)
#define BX_SlowIO		9(%bx)

#define SIREG 0x1800
#define SI_CSDN_IO		(%si)
	
		/* Print message string */
#define MSG(x)	movw $(x), %si; call message

	.file	"stage1_doc.S"

	.text

	/* Tell GAS to generate 16-bit instructions so that this code works
	   in real mode. */
	.code16

.globl _start; _start:
	/*
	 * _start is loaded at 0x20000 by the DiskOnChip IPL,
	 *	and is jumped to with CS:IP 0x2000:0
	 */

	
	/* What we need to do is:
	   1. Check the current end-of-memory in the BIOS
	   2. Reduce it by the amount of memory we need for our INT 18h 
	   3. Return, and wait for our INT 18h to be called.
	*/

		/* Store the DiskOnChip segment */
	movw	%ds, %cs:doc_seg
		
		/* Find top of memory */
	xorw	%ax, %ax
	movw	%ax, %ds
	movw	0x0413, %ax

		/* Set up our shiny net INT 18h handler */
	movw	%es,0x0066
	movw	$int18,0x0062


		/* Steal 1K from the top */
	decw	%ax
	movw	%ax,0x0413

		/* Generate segment address */
	shlw	$6,%ax
	movw	%ax,%es

		/* Debug: print the address at which we're installing */
	pushw	%cs
	popw	%ds

		/* Copy ourself into the new segment we've just reserved */
	movw	$0x80, %cx
	cld
	xorw	%si,%si
	xorw	%di,%di
	rep
	movsw

#if 0
		/* Copy the old INT 18h handler vector */
	xorw	%ax, %ax
	movw	%ax, %ds

	movw	0x0060, %ax
	movw	%ax, %es:oldvec_ofs

	movw	0x0062, %ax
	movw	%ax, %es:oldvec_seg

		/* Set up our shiny net INT 18h handler */
	//sti
	movw	%es,0x0062
	movw	$int18,0x0060
	//cli
#endif
	lret	
	
int18:
#if 0
	pushw	%ax
	pushw	%bx
	pushw	%cx
	pushw	%dx
	pushw	%ds
	pushw	%es
	pushw	%di
	pushw	%si
#endif	
	pushw	%cs
	popw	%ds

		/* Say hello */
	MSG(int18_execed_string)

		/* Check where the DiskOnChip is */
	movw	doc_seg, %ds

		/* Set up our target address for writing stage2 */
	movw	$GRUBLOADSEG, %ax
	movw	%ax, %es
#if	GRUBLOADOFS == 0
	xorw	%di, %di
#else
	movw	$GRUBLOADOFS, %di
#endif

		/* Contents of DiskOnChip are loaded at 0000:8000 */
		/* The first 0x200 bytes of this is in fact this bootloader,
		 *   and it replaces the bootloader which CVS grub has now
		 *   added to the beginning of stage2
		 * Stage2 proper starts at 0x08200, which is convenient.
		 */

	cld

		/* Copy it.... */

		/* Reset DiskOnChip */
	movw	$BXREG, %bx
	movb	$DOC_MODE_CLR_ERR + DOC_MODE_MDWREN + DOC_MODE_NORMAL, BX_DOCControl
	movb	$DOC_MODE_CLR_ERR + DOC_MODE_MDWREN + DOC_MODE_NORMAL, BX_DOCControl

	movw	$SIREG, %si
	
		/* Flash command:	Reset */
	movb	$NAND_CMD_RESET, %ah
	call	doc_cmd

		/* Load %CX with the number of 256-byte blocks to load */	
	movw	$LOADLEN, %cx

mainloop:
		/* We're still reading a block */
//	testw	$0xff, %di
//	jnz	readbyte

		/* Send new READ command - is it READ0 or READ1? */

	movw	%di, %ax
	andb	$1,%ah
	call	doc_cmd

		/* Start of new block. Set address */
	movb	$CSDN_CTRL_FLASH_IO + CSDN_CTRL_WP + CSDN_CTRL_ALE + CSDN_CTRL_CE, BX_CSDNControl
	incw	BX_ChipID

		/* We sent bits 0-7, then 9-16, then 17-23 */
		/* Yes bit 8 is missing. Read the NAND flash specs */		
	movw	%di,%dx
		/* Bits 0-7 are always zero */
	movb	$0,BX_SlowIO
	movb	$0,SI_CSDN_IO
			/* Bits 9-16 */
	shrw	$1,%dx
	.byte	0x80, 0xce /* orb adrbit16, %dh */
adrbit16:	
	.byte	0	
	movb	%dh,BX_SlowIO
	movb	%dh,SI_CSDN_IO
		/* Bits 17-23 are also zero */
	.byte	0xb2 /* movb adrbytehi, %dl */
adrbytehi:
	.byte	0
	movb	%dl,BX_SlowIO
	movb	%dl,SI_CSDN_IO

		/* Clear the ALE line to the flash chip */
	movb	$CSDN_CTRL_FLASH_IO + CSDN_CTRL_WP + CSDN_CTRL_CE, BX_CSDNControl
	call	doc_wait

	pushw	%cx		/* Store the 'blocks remaining' count */
	movw	$0x100, %cx	/* Set up to copy 0x100 bytes */
readbyte:	
	testb	BX_SlowIO,%al	/* dummy read */
	movsb			/* movb SI_CSDN_IO ;  stosb would be more obvious, */
	decw	%si		/*  but would take an extra byte. */
	loop	readbyte


	testw	$0xffff, %di	/* Check if we've done a whole 0x10000 bytes */
	jnz	endloop		/* No - continue regardless */
		/* Yes - increase %es */
	movw	%es, %cx
	addb	$0x10, %ch
	movw	%cx, %es

	addb	$0x80, %cs:adrbit16	/* Increase bit 16 */
	jnc	endloop			/* Did it overflow? */
	incb	%cs:adrbytehi		/* If so, increase the high byte too */

endloop:	
	popw	%cx		/* Restore the 'blocks remaining' count */
	loop	mainloop	/* Loop till completely done */
good:	
		/* Run it:	 jmpf 0:2000 */

	.byte	0xea
	.word	GRUBSTART,0

		/* doc_cmd:	 Send a command to the flash chip */
doc_cmd:
		/* Enable CLE line to flash */
	movb	$CSDN_CTRL_FLASH_IO + CSDN_CTRL_WP + CSDN_CTRL_CLE + CSDN_CTRL_CE, BX_CSDNControl
		/* Dummy */
	incw	-4(%bx)
		/* Write the actual command */
	movb	%ah,BX_SlowIO
	movb	%ah,(%si)

		/* doc_wait:	 Wait for the DiskOnChip to be ready */
doc_wait:
	incw	BX_ChipID
l38:	
	testb	$0x80,BX_CSDNControl
	jz	l38
	test	BX_CSDNControl,%dx
	ret
		
#if 0	
bad:		/* Whinge and return*/
	
	popw	%si
	popw	%di
	popw	%es
	popw	%ds
	popw	%dx
	popw	%cx
	popw	%bx
	popw	%ax
	.byte 0xea		// jmpf
oldvec_ofs:	.word 0
oldvec_seg:	.word 0
#endif

doc_seg:	.word 0
	
int18_execed_string:	.string "Loading GRUB from DiskOnChip\n\r"

/*
 * message: write the string pointed to by %si
 *
 *   WARNING: trashes %si, %ax, and %bx
 */

	/*
	 * Use BIOS "int 10H Function 0Eh" to write character in teletype mode
	 *	%ah = 0xe	%al = character
	 *	%bh = page	%bl = foreground color (graphics modes)
	 */
1:
	movw	$0x0001, %bx
	movb	$0xe, %ah
	int	$0x10		/* display a byte */

message:
	lodsb
	cmpb	$0, %al
	jne	1b	/* if not end of string, jmp to display */
	ret

	/* Hex output routines, used at one point in debugging */	
#if 0
phword:
	pushw	%ax
	xchgb	%al,%ah
	call	phbyte
	movb	%ah,%al
	call	phbyte
	popw	%ax
	ret


phbyte:
	pushw	%ax
	movb	%al, %ah
	shrb	$4,%al
	call	phnibble
	movb	%ah, %al
	call	phnibble
	popw	%ax
	ret

phnibble:
	pushw	%ax
	andb	$0xf,%al
	addb	$48,%al
	cmpb	$57,%al
	jna	ph1
	add	$7,%al
ph1:	mov	$0xe,%ah
	int	$0x10
	popw	%ax
	ret
#endif

#if 0
		/* Cause an error if this code was too large. */
	.org 0xff
	.byte 0 /* Placement of checksum byte */
#endif	
