#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <errno.h>

#include <asm/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nftl.h>

unsigned char BadUnitTable[512];
unsigned char readbuf[512];
unsigned char writebuf[4][512];

mtd_info_t meminfo;
erase_info_t erase;
unsigned long retlen, ofs;
unsigned long startofs = 0, size = 0;
unsigned long ezones = 0, ezone=0;
int fd;
long MediaUnit1=-1, MediaUnit2=-1;
struct NFTLMediaHeader *NFTLhdr;

static inline unsigned char checkblock(unsigned long block) 
{
	unsigned long ofs = block * meminfo.erasesize;

#ifdef CHECK_FLASH	
	unsigned long i;
	unsigned long blockofs;

	/* First erase the block */
	
	erase.start = ofs;

	if (ioctl(fd,MEMERASE,&erase) != 0) {      
		return ZONE_BAD_ORIGINAL;
	}
	   
	for (blockofs=0; blockofs < meminfo.erasesize; blockofs += 512) {
		pread(fd, readbuf, 512, ofs + blockofs);

		if (memcmp(readbuf, writebuf[0], 512)) {
			   /* Block wasn't 0xff after erase */
			printf("Block not 0xff after erase\n");
			   return ZONE_BAD_ORIGINAL;
		}

		pwrite(fd, writebuf[1], 512, blockofs + ofs);
		pread(fd, readbuf, 512, blockofs + ofs);

		if (memcmp(readbuf, writebuf[1], 512)) {
			printf("Block not zero after clearing\n");
			return ZONE_BAD_ORIGINAL;
		}
	}
	
	if (ioctl(fd,MEMERASE,&erase) != 0) {      
		return ZONE_BAD_ORIGINAL;
	}
	
	for (blockofs=0; blockofs < meminfo.erasesize; blockofs += 512) {
		
		pwrite(fd, writebuf[2], 512, blockofs + ofs);
		pread(fd, readbuf, 512, blockofs + ofs);

		if (memcmp(readbuf, writebuf[2], 512)) {
			printf("Block not 0x5a after writing\n");
			return ZONE_BAD_ORIGINAL;
		}
	}

	if (ioctl(fd,MEMERASE,&erase) != 0) {      
		return ZONE_BAD_ORIGINAL;
	}
	   
	for (blockofs=0; blockofs < meminfo.erasesize; blockofs += 512) {

		pwrite(fd, writebuf[3], 512, blockofs + ofs);
		pread(fd, readbuf, 512, blockofs + ofs);

		if (memcmp(readbuf, writebuf[3], 512)) {
			printf("Block not 0xa5 after writing\n");
			return ZONE_BAD_ORIGINAL;
		}
	}
	//printf("block %lx good\n", block);

#endif
	erase.start = ofs;
	
	if (ioctl(fd,MEMERASE,&erase) != 0) {      
		return ZONE_BAD_ORIGINAL;
	}
	return ZONE_GOOD;

}


int main(int argc,char **argv)
{
	unsigned char oobbuf[16];
	struct mtd_oob_buf oob = {0, 16, oobbuf};

   if (argc < 2) {
	   fprintf(stderr,"Usage: %s <mtddevice> [<start offset> [<size>]]\n");
	   return 1;
   }
   
   if (argc > 2) {
	   startofs = atoi(argv[2]);
   }
   if (argc > 3) {
	   size = atoi(argv[2]);
   }
	   
   // Open and size the device
   if ((fd = open(argv[1],O_RDWR)) < 0) {
	   perror("Open flash device");
	   return 1;
   }
   
   if (ioctl(fd,MEMGETINFO,&meminfo) != 0) {
	   perror("ioctl(MEMGETINFO)");
	   close(fd);
	   return 1;
   }
   
   if (meminfo.erasesize != 0x2000) {
	   printf("Erase size not 8Kb - I'm confused\n");
	   close(fd);
	   return 1;
   }

   memset(writebuf[0], 0xff, 512);
   memset(writebuf[1], 0x00, 512);
   memset(writebuf[2], 0x5a, 512);
   memset(writebuf[3], 0xa5, 512);

   if (!size) size = meminfo.size - startofs;

   erase.length = meminfo.erasesize;

   ezones = size / meminfo.erasesize;
   
   if ( ezones > MAX_ERASE_ZONES ) {
	   /* Ought to change the UnitSizeFactor. But later. */
	   size = meminfo.erasesize * MAX_ERASE_ZONES;
	   ezones = MAX_ERASE_ZONES;
   }
   
   for (ezone = startofs / meminfo.erasesize; 
	ezone < ( ezones + startofs / meminfo.erasesize); 
	ezone++) {
	   int someflagged;

	   if (!(ezone % 512)) {
		   /* New block of zones */
		   memset(BadUnitTable, 0xff, 512);
		   someflagged = 0;
	   }
	   
	   if ((BadUnitTable[ezone % 512] = checkblock(ezone)) == ZONE_GOOD) {
		   if (MediaUnit1 == -1) {
			   MediaUnit1 = ezone;
		   } else if (MediaUnit2 == -1) {
			   MediaUnit2 = ezone;
		   }
	   }
	   else {
		   someflagged = 1;
	   }
			  
	   if (ezone & 0x1ff == 0x1ff) {
		   /* We have a whole chunk to write. */
		   if (MediaUnit1 == -1 || MediaUnit2 == -1) {
			   printf("Need to write data, but don't yet know MediaUnit - WTF?\n");
			   return 1;
		   }
		   
		   if (someflagged) { 
			   /* Only need to write data if there are bad blocks */
			   pwrite(fd, BadUnitTable, 512, 
				  (MediaUnit1 * meminfo.erasesize) + 512 + 
				  (ezone % 512));

			   pwrite(fd, BadUnitTable, 512, 
				  (MediaUnit2 * meminfo.erasesize) + 512 + 
				  (ezone % 512));
		   }
		   
	   }
   }

   /* FIXME: We won't actually deal with that info - it's always clean so 
      far and I'm only doing this so I don't have to keep booting DOS to
      fix the NFTL when I break it :) 
   */
   
   NFTLhdr = (struct NFTLMediaHeader *)(writebuf[0]);

   strcpy(NFTLhdr->DataOrgID, "ANAND");

   size -= (2*meminfo.erasesize);
   //   startofs += ( 2*meminfo.erasesize);

   NFTLhdr->NumEraseUnits = (size / meminfo.erasesize);
   NFTLhdr->FirstPhysicalEUN = MediaUnit2 + 1;
   NFTLhdr->FormattedSize = size - 0x4000;
   
   pwrite(fd, writebuf[0], 512, MediaUnit1 * meminfo.erasesize);
   pwrite(fd, writebuf[0], 512, MediaUnit2 * meminfo.erasesize);

   memset(oobbuf, 0xff, 16);
   oobbuf[11] = oobbuf[10] = oobbuf[9] = 0;
   oobbuf[8] = 3;
   oobbuf[12] = oobbuf[14] = 0x69;
   oobbuf[13] = oobbuf[15] = 0x3c;
   
   for (ezone = NFTLhdr->FirstPhysicalEUN ; 
	ezone < ( ezones + startofs / meminfo.erasesize); ezone++) {
	   int someflagged;

	   oob.start = (ezone * meminfo.erasesize) + 512;
	   if (	   ioctl(fd, MEMWRITEOOB, &oob))
		   printf("MEMWRITEOOB at %lx: %s\n",oob.start, sys_errlist[errno]);
	   
   }
}
